#lang racket
;;; PF2 - Automne 2016 - Racket 6.6
;;; sol2.rkt

(require rackunit        ; pour la macro check-equal?
         "utils.rkt")    ; pour les macros while, show et quelques autres gâteries...

; CONSULTEZ CET ARTICLE DE LA WIKIPEDIA, NOTAMMENT LA SECTION `Scheme' :
; http://fr.wikipedia.org/wiki/Fermeture_(informatique)

;;; ==================== TP2 sols ======================

"Exo 2.2"

; somme des carrés des impairs de [1,100] en fonctionnel PF1 à l'ordre supérieur (one-liner) :
(printf "fonctionnel ordre sup : 1^2 + 3^2 + ... + 99^2 = ~a\n"
        (apply + (map sqr (filter odd? (build-list 100 add1)))))

; le même en impératif avec une boucle while
(printf "imperatif-while       : 1^2 + 3^2 + ... + 99^2 = ~a\n"
        (let ((s 0) (k 1))
          (while (< k 100)
            (set! s (+ s (sqr k)))
            (set! k (+ k 2)))
          s))

; le même en impératif avec une boucle for en style Python
(printf "imperatif-for         : 1^2 + 3^2 + ... + 99^2 = ~a\n"
        (let ((s 0))
          (for ([k (in-range 1 100 2)])      ; for/product pour calculer un produit (n'existe pas en Python)
            (set! s (+ s (sqr k))))
          s))

; le mzme en impératif avec une boucle for/sum de sommation (analogue de sum en Python)
(printf "imperatif-for/sum     : 1^2 + 3^2 + ... + 99^2 = ~a\n"
        (for/sum ([k (in-range 1 100 2)])      ; for/product pour calculer un produit (n'existe pas en Python)
          (sqr k)))

; le même en fonctionnel récursif PF1 pour ne pas perdre la main :-)

(define (sci a b)   ; a et b impairs
  (if (> a b)
      0
      (+ (sqr a) (sci (+ a 2) b))))

(printf "fonctionnel pur       : 1^2 + 3^2 + ... + 99^2 = ~a\n" (sci 1 99))

(define (reverse_v1 L)      ; while impératif, temps linéaire
  (define acc '())          ; ou bien avec un 'let' old style...
  (while (not (null? L))
    (set! acc (cons (car L) acc))
    (set! L (cdr L)))
  acc)

(check-equal? (reverse_v1 '(fa (re mi) fa sol)) '(sol fa (re mi) fa))   ; check-equal? dans le module rackunit

(define (reverse_v2 L)     ; avec la boucle for de Racket
  (define res '())
  (for ([x (in-list L)])   ; ou bien :    for ([x L])...
    (set! res (cons x res)))
  res)

(check-equal? (reverse_v2 '(a z e r t y)) (reverse '(a z e r t y)))

"Exo 2.3"

(define-syntax empiler!
  (syntax-rules ()
    ((empiler! x pile) (set! pile (cons x pile)))))                         ; aucun resultat

(define-syntax depiler!
  (syntax-rules ()
    ((depiler! pile) (let ((x (car pile))) (set! pile (cdr pile)) x))))     ; avec resultat (begin implicite dans le corps du let)

"Exo 2.4"

(define-syntax loop    ; existe dans le Scheme du logiciel musical Grace :  http://commonmusic.sourceforge.net
  (syntax-rules (For From To By Sum In When Collect)
    ((loop For i From a To b By s Sum expr) 
     (apply + (map (lambda (i) expr) (range a (+ b 1) s))))   ; <==> (for/sum ([i (in-range a (+ b 1) s)]) expr))
    ((loop For x In L When test Collect expr) 
     (map (lambda (x) expr) (filter (lambda(x) test) L)))))   ; <==> (for/list ([x L] #:when test) expr)))) 

(check-equal? (loop For i From 1 To 10 By 2 Sum (sqr i)) 165)
(check-equal? (loop For x In '(3 2 5 4 9) When (even? x) Collect x) '(2 4))

"Exo 2.5"

(define (make-gen-fac)
  (define n 0)
  (define f 1)                 ; les neurones de la fermeture, f = n!
  (lambda ()
    (define current-f f)       ; sauvegarder la valeur courante pour ne pas perdre le premier...
    (set! n (+ n 1))
    (set! f (* n f))
    current-f))

(define GF (make-gen-fac))     ; un générateur de factielles

(printf "factorielles avec generateur : ")
(for ([i (in-range 11)])       ; affichage de 0! 1! ... 10!
  (printf "~a " (GF)))
(printf "...\n")

(define (make-gen-random-distincts)   ; générateur de nombres aléatoires distincts de [-99,99]
  (define L '())
  (define (generate)
    (define x (* (random 100) (if (= (random 2) 0) 1 -1)))
    (if (member x L)
        (generate)
        (begin (set! L (cons x L)) x)))
  generate)      ; on rend une fonction d'arité 0

(define GAD (make-gen-random-distincts))
(printf "générateur de 10 nombres aléatoires distincts : ")
(for/list ([i (in-range 10)]) (GAD))

"Exo 2.6"

(define AL-binomial              ; avec A-listes, un peu lent car mémoire scannée en O(n)
  (let ((AL '()))
    (define (calcul n p)
      (if (or (= n p) (= p 0))
          1
          (let ((essai (assoc (list n p) AL)))    ; deja en memoire ?
            (if essai
                (cadr essai)
                (let ((res (+ (calcul (- n 1) p) (calcul (- n 1) (- p 1)))))
                  (set! AL (cons (list (list n p) res) AL))  ; on le stocke
                  res)))))
    calcul))      ; on rend une fonction d'arité 2

(printf "(AL-binomial 1000 20) lent (une quinzaine de sec) : ")
(time (AL-binomial 1000 20))          ; plusieurs secondes, lent...

(define H1-binomial          ; avec table de hash-code mutable, les dictionnaires mutables de Python
  (let ((H (make-hash)))     ; beaucoup plus rapide qu'une A-liste !
    (define (calcul n p)
      (cond ((or (= n p) (= p 0)) 1)
            ((hash-has-key? H (list n p)) (hash-ref H (list n p)))
            (else (let ((v (+ (calcul (- n 1) p) (calcul (- n 1) (- p 1)))))
                    (hash-set! H (list n p) v)
                    v))))
    calcul))

(printf "(H1-binomial 1000 20) rapide :\n")
(time (H1-binomial 1000 20))            ; fast !
(time (H1-binomial 1000 20))            ; faster : 0 sec

; Utilisation d'une table de hash FONCTIONNELLE (non mutable) ! Je dois la passer en paramètre puisqu'elle
; n'est pas mutable, ok ? Oui mais il faut alors que je rende le résultat du calcul mais en même temps la
; table de hash étendue, donc fonction à deux resultats (res new-H)
; Bon, ok, c'est un peu plus lourd et moins rapide mais ca fait bien travailler les passages de paramètres :-)

(define (H2-binomial n p H)             ; on rend (res new-H)
  (cond ((or (= n p) (= p 0)) (list 1 H))
        ((hash-has-key? H (list n p)) (list (hash-ref H (list n p)) H))   ; H n'a pas changee
        (else (let* ((res1 (H2-binomial (- n 1) p H))
                     (v1 (car res1))
                     (new-H1 (cadr res1))
                     (res2 (H2-binomial (- n 1) (- p 1) new-H1))
                     (v2 (car res2))
                     (new-H2 (cadr res2))
                     (res (+ v1 v2)))
                (list res (hash-set new-H2 (list n p) res))))))

(define essai1 (H2-binomial 5 2 (hash)))
(printf "L'appel (H2-binomial 5 2 (hash)) rend deux résultats, le binomial ~a et la table de hash construite :\n~a\n"
        (car essai1) (cadr essai1))
(printf "Voici (H2-binomial 1000 20 (hash)) rapide exécuté deux fois :\n")
(define essai2 (time (H2-binomial 1000 20 (hash))))
(time (car (H2-binomial 1000 20 (cadr essai2))))            ; (cadr essai2) est la table de hash resultat de essai2

; affichage de la ligne 10 du triangle de Pascal : les (H-binomial 10 p) avec 0 <= p <= 10

(printf "ligne 10 du triangle de Pascal : ")
(for/list ([p (in-range 11)])
  (H1-binomial 10 p))

"Exo 2.7"
; Fibonacci itératif. On déplace n fois un plateau (a,b) de deux termes consécutifs de la suite de Fibonacci
(define (fib n)       ; en O(n)                 ; def fib(n) :
  (define a 0)                                  ;     a = 0
  (define b 1)                                  ;     b = 1
  (for ([i (in-range n)])                       ;     for i in range(n) :
    (define old-a a)                            ;         (a,b) = (b,a+b)      # en Python ! Cool...
    (set! a b)                                  ;         # il faudrait donc définir en Scheme
    (set! b (+ old-a b)))                       ;         # une macro (affect x y u v) réalisant le (x,y) = (u,v) de Python
  a)                                            ;     return a

(check-equal? (fib 20) 6765)

"Exo 2.8"

(define ($for-each proc L)    ; rien de mieux que programmer pour comprendre :-)
  (for ([x L]) (proc x)))     ; personne n'oserait écrire (void (map proc L)), ok ?

(printf "for-each  : ")
(for-each print '(a b c))     ; <==> (for ([x '(a b c)]) (print x))
(printf "\n$for-each : ")
($for-each print '(a b c))
(printf "\n")

(define ($$reverse L)
  (define res '())
  (for-each (lambda (x) (empiler! x res)) L)  ; <==> (for ([x L]) (empiler! x res))
  res)

(check-equal? ($$reverse '(a (b c) d)) (reverse '(a (b c) d)))

"Exo 2.9"

(define (make-gen-fib)     ; retourne un générateur des termes de la suite de Fibonacci
  ; meme stratégie que l'exo 4.7 mais on ne fait qu'un pas et non une itération
  (define a 0)
  (define b 1)
  (lambda ()
    (let ((old-a a))
      (set! a b)
      (set! b (+ old-a b))
      old-a)))

(define fibo (make-gen-fib))
(check-equal? (for/list ([i (in-range 10)]) (fibo))      ; le générateur
              (for/list ([i (in-range 10)]) (fib i)))    ; exo 4.7

"Exo 2.10"

;;; animation dans un monde mutable (avec affectations)

(require 2htdp/universe 2htdp/image)

(define (anim-balle)    ; une balle qui rebondit à la verticale dans un tube
  (define BALLE (circle 30 'solid "red"))
  (define LARG 80) (define LARG/2 (quotient LARG 2))
  (define HAUT 300) (define HAUT/2 (quotient HAUT 2))
  (define G 1)                     ; gravitation
  (define F 0.96)                  ; coefficient de frottement au sol (perte d'energie)
  (define y (quotient HAUT 2))     ; ordonnée du point bas de la balle
  (define dy 0)                    ; vitesse de la balle
  (define FOND (place-image (text (format "F=~a" F) 18 "blue") LARG/2 20 (rectangle LARG HAUT 'solid "yellow")))
  (define (suivant ?)              ; Le monde m est ici bidon, on ne s'en sert pas, on fait directement muter les variables (brrr)
    (set! y (+ y dy))
    (when (> (+ y 30) HAUT)
      (begin (set! y (- HAUT 30))
             (set! dy (* F (- dy)))))   ; frottement au sol
    (set! dy (+ dy G)))
  (define (dessiner ?)             ; Le monde m est ici bidon, on ne s'en sert pas
    (place-image BALLE LARG/2 y FOND))
  (define (final? ?)              ; Le monde m est ici bidon, on ne s'en sert pas
    (and (< (abs (- y -30 HAUT)) 0.1) (< (abs dy) 1))) 
  (big-bang (void)                ; Le monde initial est bidon...
            (on-tick suivant)
            (on-draw dessiner)
            (stop-when final?)))

(anim-balle)

(require racket/system)
; system permet de lancer une ligne de commande à partir de Racket.
; open est une commande Unix standard sur Mac, mais qu'il vous faudra peut-être adapter. Elle ouvre html, pdf, txt, rkt, py, etc.
; Rappel : l'Unix du Mac est une variante de Unix-BSD alors que Linux dérive de Unix-System-V.
(void (case (system-type)
        ((macosx) (system "open https://www.freebsd.org/doc/fr/articles/explaining-bsd/index.html"))
        ((windows) (system "start https://www.freebsd.org/doc/fr/articles/explaining-bsd/index.html"))
        (else '?)))

; Si vous trouvez un bon équivalent de `open' pour Linux, adaptez et envoyez un mail, merci :-)
; On peut tester l'O.S. avec la fonction Racket (system-type)



